/*
 * The OpenSAML License, Version 1.
 * Copyright (c) 2002
 * University Corporation for Advanced Internet Development, Inc.
 * All rights reserved
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution, if any, must include
 * the following acknowledgment: "This product includes software developed by
 * the University Corporation for Advanced Internet Development
 * <http://www.ucaid.edu>Internet2 Project. Alternately, this acknowledegement
 * may appear in the software itself, if and wherever such third-party
 * acknowledgments normally appear.
 *
 * Neither the name of OpenSAML nor the names of its contributors, nor
 * Internet2, nor the University Corporation for Advanced Internet Development,
 * Inc., nor UCAID may be used to endorse or promote products derived from this
 * software without specific prior written permission. For written permission,
 * please contact opensaml@opensaml.org
 *
 * Products derived from this software may not be called OpenSAML, Internet2,
 * UCAID, or the University Corporation for Advanced Internet Development, nor
 * may OpenSAML appear in their name, without prior written permission of the
 * University Corporation for Advanced Internet Development.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND WITH ALL FAULTS. ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, AND NON-INFRINGEMENT ARE DISCLAIMED AND THE ENTIRE RISK
 * OF SATISFACTORY QUALITY, PERFORMANCE, ACCURACY, AND EFFORT IS WITH LICENSEE.
 * IN NO EVENT SHALL THE COPYRIGHT OWNER, CONTRIBUTORS OR THE UNIVERSITY
 * CORPORATION FOR ADVANCED INTERNET DEVELOPMENT, INC. BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/* SAMLAttribute.cpp - abstract SAML attribute class

   Scott Cantor
   2/23/02
*/

#include "internal.h"

using namespace saml;
using namespace std;

SAMLAttributeFactory* SAMLAttribute::m_factory=NULL;

SAMLAttributeFactory* SAMLAttribute::setFactory(SAMLAttributeFactory* factory)
{
    SAMLConfig::getConfig().saml_lock();
    SAMLAttributeFactory* ret=m_factory;
    m_factory=factory;
    SAMLConfig::getConfig().saml_unlock();
    return ret;
}

SAMLAttribute* SAMLAttribute::getInstance(DOMElement* e)
{
    if (!e)
        throw MalformedException(SAMLException::RESPONDER,"SAMLAttribute::getInstance given an empty DOM");

    SAMLAttribute* a=NULL;
    SAMLConfig::getConfig().saml_lock();
    SAMLAttributeFactory* f=m_factory;
    SAMLConfig::getConfig().saml_unlock();
    return (f) ? f(e) : new SAMLAttribute(e);
}

SAMLAttribute* SAMLAttribute::getInstance(istream& in)
{
    XML::Parser p;
    XML::StreamInputSource src(in);
    Wrapper4InputSource dsrc(&src,false);
    DOMDocument* doc=p.parse(dsrc);
    try
    {
        SAMLAttribute* a=getInstance(doc->getDocumentElement());
        a->setDocument(doc);
        return a;
    }
    catch(...)
    {
        doc->release();
        throw;
    }
}

void SAMLAttribute::valueToDOM(unsigned int index, DOMElement* e) const
{
    e->appendChild(e->getOwnerDocument()->createTextNode(m_values[index]));
}

SAMLAttribute::SAMLAttribute(
    const XMLCh* name, const XMLCh* ns, const saml::QName* type, long lifetime, const Iterator<const XMLCh*>& values
    ) : m_name(XMLString::replicate(name)), m_namespace(XMLString::replicate(ns)), m_lifetime(lifetime)
{
    RTTI(SAMLAttribute);
    m_type=type ? new saml::QName(*type) : NULL;
    while (values.hasNext())
        m_values.push_back(XMLString::replicate(values.next()));
}

SAMLAttribute::SAMLAttribute(DOMElement* e) : m_name(NULL), m_namespace(NULL), m_type(NULL), m_lifetime(0)
{
    RTTI(SAMLAttribute);
    fromDOM(e);
}

SAMLAttribute::SAMLAttribute(istream& in) : SAMLObject(in), m_name(NULL), m_namespace(NULL), m_type(NULL)
{
    RTTI(SAMLAttribute);
    fromDOM(m_document->getDocumentElement());
}

SAMLAttribute::~SAMLAttribute()
{
    if (m_bOwnStrings)
    {
        XMLString::release(const_cast<XMLCh**>(&m_name));
        XMLString::release(const_cast<XMLCh**>(&m_namespace));
        for (vector<const XMLCh*>::iterator i=m_values.begin(); i!=m_values.end(); i++)
        {
            XMLCh* p = const_cast<XMLCh*>(*i);
            XMLString::release(&p);
        }
    }
    delete m_type;
}

void SAMLAttribute::checkValidity() const
{
    if (!m_name || !m_namespace || m_values.size()==0)
        throw MalformedException(SAMLException::RESPONDER, "Attribute invalid, requires name and namespace, and at least one value");
}

SAMLObject* SAMLAttribute::clone() const
{
    return new SAMLAttribute(m_name,m_namespace,m_type,m_lifetime,m_values);
}

void SAMLAttribute::fromDOM(DOMElement* e)
{
    saml::NDC ndc("fromDOM");
    SAMLObject::fromDOM(e);
    if (SAMLConfig::getConfig().strict_dom_checking &&
        (XMLString::compareString(XML::SAML_NS,e->getNamespaceURI()) || XMLString::compareString(L(Attribute),e->getLocalName())))
            throw MalformedException(SAMLException::RESPONDER,"SAMLAttribute::fromDOM() missing saml:Attribute element at root");

    m_root=e;
    m_bOwnStrings=false;
    m_name=e->getAttributeNS(NULL,L(AttributeName));
    m_namespace=e->getAttributeNS(NULL,L(AttributeNamespace));

    e=XML::getFirstChildElement(e,XML::SAML_NS,L(AttributeValue));
    while (e)
    {
        if (!m_type)
            m_type=saml::QName::getQNameAttribute(e,XML::XSI_NS,L(type));

        const XMLCh* ptr=NULL;
        DOMNode* n=e->getFirstChild();
        if (n && n->getNodeType()==DOMNode::TEXT_NODE && (ptr=n->getNodeValue()))
            m_values.push_back(ptr);
        else {
            m_values.push_back(&chNull);
            saml::NDC ndc("addValue");
            SAML_log.warn("ignoring an AttributeValue without a single, non-empty text node");
        }
        
        m_elements.push_back(e);
        e=XML::getNextSiblingElement(e,XML::SAML_NS,L(AttributeValue));
    }
}

void SAMLAttribute::setName(const XMLCh* name)
{
    if (!name || !*name)
        throw MalformedException("name cannot be null");
    
    if (m_root) {
        DOMAttr* a=static_cast<DOMElement*>(m_root)->getAttributeNodeNS(NULL,L(AttributeName));
        a->setNodeValue(name);
        if (!m_bOwnStrings)
            m_name=a->getNodeValue();
    }

    if (m_bOwnStrings) {
        XMLString::release(const_cast<XMLCh**>(&m_name));
        m_name=XMLString::replicate(name);
    }
}

void SAMLAttribute::setNamespace(const XMLCh* ns)
{
    if (!ns || !*ns)
        throw MalformedException("namespace cannot be null");
    
    if (m_root) {
        DOMAttr* a=static_cast<DOMElement*>(m_root)->getAttributeNodeNS(NULL,L(AttributeNamespace));
        a->setNodeValue(ns);
        if (!m_bOwnStrings)
            m_namespace=a->getNodeValue();
    }

    if (m_bOwnStrings) {
        XMLString::release(const_cast<XMLCh**>(&m_namespace));
        m_name=XMLString::replicate(ns);
    }
}

XMLCh* SAMLAttribute::computeTypeDecl(DOMElement* e) const
{
    static const XMLCh sep[]={chColon, chNull};
    XMLCh* xsitype=NULL;
    e->removeAttributeNS(XML::XMLNS_NS,L_QNAME(xmlns,typens));
    if (m_type) {
        const XMLCh* prefix=NULL;
        if (!XMLString::compareString(XML::XSD_NS,m_type->getNamespaceURI()))
            prefix=L(xsd);
        else {
            e->setAttributeNS(XML::XMLNS_NS,L_QNAME(xmlns,typens),m_type->getNamespaceURI());
            prefix=L(typens);
        }
        xsitype=new XMLCh[XMLString::stringLen(prefix) + XMLString::stringLen(m_type->getLocalName()) + 2];
        XMLString::catString(xsitype,prefix);
        XMLString::catString(xsitype,sep);
        XMLString::catString(xsitype,m_type->getLocalName());
    }
    return xsitype;
}

void SAMLAttribute::setType(const QName* type)
{
    delete m_type;
    if (type)
        m_type=new QName(*type);
    if (m_root) {
        XMLCh* xsitype=computeTypeDecl(static_cast<DOMElement*>(m_root));
        DOMElement* child=XML::getFirstChildElement(m_root);
        while (child) {
            child->removeAttributeNS(XML::XSI_NS,L_QNAME(xsi,type));
            if (xsitype)
                child->setAttributeNS(XML::XSI_NS,L_QNAME(xsi,type),xsitype);
            child=XML::getNextSiblingElement(child);
        }
        delete[] xsitype;
    }
}

void SAMLAttribute::setLifetime(long lifetime)
{
    m_lifetime=lifetime;
}

void SAMLAttribute::setValues(const Iterator<const XMLCh*>& values)
{
    while (m_values.size())
        removeValue(0);
    while (values.hasNext())
        addValue(values.next());
}

void SAMLAttribute::addValue(const XMLCh* value)
{
    if (value) {
        if (m_root) {
            auto_ptr_XMLCh xsitype(computeTypeDecl(static_cast<DOMElement*>(m_root)));
            DOMElement* v=m_root->getOwnerDocument()->createElementNS(XML::SAML_NS, L(AttributeValue));
            if (xsitype.get())
                v->setAttributeNS(XML::XSI_NS,L_QNAME(xsi,type),xsitype.get());
            
            m_values.push_back(value);
            valueToDOM(m_values.size()-1,v);
            m_root->appendChild(v);
            m_elements.push_back(v);
            m_values.erase(m_values.end()-1);
            if (!m_bOwnStrings)
                m_values.push_back(v->getFirstChild()->getNodeValue());
        }
        
        if (m_bOwnStrings)
            m_values.push_back(XMLString::replicate(value));
            
        if (!m_sbValues.empty()) {
            char* temp=toUTF8(value);
            m_sbValues.push_back(temp);
            delete[] temp;
        }
    }
    else
        throw MalformedException("value cannot be null");
}

void SAMLAttribute::removeValue(unsigned long index)
{
    if (m_bOwnStrings) {
        XMLCh* p=const_cast<XMLCh*>(m_values[index]);
        XMLString::release(&p);
    }
    m_values.erase(m_values.begin()+index);
    if (!m_sbValues.empty())
        m_sbValues.erase(m_sbValues.begin()+index);
    
    if (m_root) {
        DOMElement* e = const_cast<DOMElement*>(m_elements[index]);
        if (e) {
            m_elements.erase(m_elements.begin()+index);
            m_root->removeChild(e);
            e->release();
        }
    }

    if (m_values.size()==0) {
        saml::NDC("removeValue");
        SAML_log.info("all values have been removed, attribute is in an illegal state");
    }
}

DOMNode* SAMLAttribute::toDOM(DOMDocument* doc, bool xmlns) const
{
    SAMLObject::toDOM(doc,xmlns);
    if (m_root)
    {
        if (xmlns)
        {
            static_cast<DOMElement*>(m_root)->setAttributeNS(XML::XMLNS_NS,L(xmlns),XML::SAML_NS);
            static_cast<DOMElement*>(m_root)->setAttributeNS(XML::XMLNS_NS,L_QNAME(xmlns,xsi),XML::XSI_NS);
            static_cast<DOMElement*>(m_root)->setAttributeNS(XML::XMLNS_NS,L_QNAME(xmlns,xsd),XML::XSD_NS);
        }
        return m_root;
    }
    if (!doc)
        doc=m_document;

    // Construct the Attribute.
    DOMElement* a=doc->createElementNS(XML::SAML_NS,L(Attribute));
    if (xmlns)
        a->setAttributeNS(XML::XMLNS_NS,L(xmlns),XML::SAML_NS);
    a->setAttributeNS(XML::XMLNS_NS,L_QNAME(xmlns,xsi),XML::XSI_NS);
    a->setAttributeNS(XML::XMLNS_NS,L_QNAME(xmlns,xsd),XML::XSD_NS);
    a->setAttributeNS(NULL,L(AttributeName),m_name);
    a->setAttributeNS(NULL,L(AttributeNamespace),m_namespace);

    auto_ptr_XMLCh xsitype(computeTypeDecl(a));
    for (unsigned int i=0; i<m_values.size(); i++)
    {
        DOMElement* v = doc->createElementNS(XML::SAML_NS,L(AttributeValue));
        if (xsitype.get())
            v->setAttributeNS(XML::XSI_NS, L_QNAME(xsi,type), xsitype.get());
        valueToDOM(i,v);
        a->appendChild(v);
        m_elements.push_back(v);
    }

    return m_root=a;
}

Iterator<string> SAMLAttribute::getSingleByteValues() const
{
    if (m_sbValues.empty())
    {
        for (vector<const XMLCh*>::const_iterator i=m_values.begin(); i!=m_values.end(); i++)
        {
            char* temp=toUTF8(*i);
            if (temp)
                m_sbValues.push_back(temp);
            delete[] temp;
        }
    }
    return Iterator<string>(m_sbValues);
}
