/*
 * The OpenSAML License, Version 1.
 * Copyright (c) 2002
 * University Corporation for Advanced Internet Development, Inc.
 * All rights reserved
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution, if any, must include
 * the following acknowledgment: "This product includes software developed by
 * the University Corporation for Advanced Internet Development
 * <http://www.ucaid.edu>Internet2 Project. Alternately, this acknowledegement
 * may appear in the software itself, if and wherever such third-party
 * acknowledgments normally appear.
 *
 * Neither the name of OpenSAML nor the names of its contributors, nor
 * Internet2, nor the University Corporation for Advanced Internet Development,
 * Inc., nor UCAID may be used to endorse or promote products derived from this
 * software without specific prior written permission. For written permission,
 * please contact opensaml@opensaml.org
 *
 * Products derived from this software may not be called OpenSAML, Internet2,
 * UCAID, or the University Corporation for Advanced Internet Development, nor
 * may OpenSAML appear in their name, without prior written permission of the
 * University Corporation for Advanced Internet Development.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND WITH ALL FAULTS. ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, AND NON-INFRINGEMENT ARE DISCLAIMED AND THE ENTIRE RISK
 * OF SATISFACTORY QUALITY, PERFORMANCE, ACCURACY, AND EFFORT IS WITH LICENSEE.
 * IN NO EVENT SHALL THE COPYRIGHT OWNER, CONTRIBUTORS OR THE UNIVERSITY
 * CORPORATION FOR ADVANCED INTERNET DEVELOPMENT, INC. BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/* SAMLAttributeDesignator.cpp - designates a SAML attribute in a query

   Scott Cantor
   1/15/04
*/

#include "internal.h"

using namespace saml;
using namespace std;

SAMLAttributeDesignator::SAMLAttributeDesignator(const XMLCh* name, const XMLCh* ns)
    : m_name(XMLString::replicate(name)), m_namespace(XMLString::replicate(ns))
{
    RTTI(SAMLAttributeDesignator);
}

SAMLAttributeDesignator::SAMLAttributeDesignator(DOMElement* e) : m_name(NULL), m_namespace(NULL)
{
    RTTI(SAMLAttributeDesignator);
    fromDOM(e);
}

SAMLAttributeDesignator::SAMLAttributeDesignator(istream& in) : SAMLObject(in), m_name(NULL), m_namespace(NULL)
{
    RTTI(SAMLAttributeDesignator);
    fromDOM(m_document->getDocumentElement());
}

SAMLAttributeDesignator::~SAMLAttributeDesignator()
{
    if (m_bOwnStrings)
    {
        XMLString::release(const_cast<XMLCh**>(&m_name));
        XMLString::release(const_cast<XMLCh**>(&m_namespace));
    }
}

void SAMLAttributeDesignator::checkValidity() const
{
    if (!m_name || !m_namespace)
        throw MalformedException(SAMLException::RESPONDER, "AttributeDesignator invalid, requires name and namespace");
}

SAMLObject* SAMLAttributeDesignator::clone() const
{
    return new SAMLAttributeDesignator(m_name,m_namespace);
}

void SAMLAttributeDesignator::fromDOM(DOMElement* e)
{
    saml::NDC ndc("fromDOM");
    SAMLObject::fromDOM(e);
    if (SAMLConfig::getConfig().strict_dom_checking &&
        (XMLString::compareString(XML::SAML_NS,e->getNamespaceURI()) || XMLString::compareString(L(AttributeDesignator),e->getLocalName())))
            throw MalformedException(SAMLException::RESPONDER,"SAMLAttributeDesignator::fromDOM() missing saml:AttributeDesignator element at root");

    m_root=e;
    m_bOwnStrings=false;
    m_name=e->getAttributeNS(NULL,L(AttributeName));
    m_namespace=e->getAttributeNS(NULL,L(AttributeNamespace));
}

void SAMLAttributeDesignator::setName(const XMLCh* name)
{
    if (!name || !*name)
        throw MalformedException("name cannot be null");
    
    if (m_root) {
        DOMAttr* a=static_cast<DOMElement*>(m_root)->getAttributeNodeNS(NULL,L(AttributeName));
        a->setNodeValue(name);
        if (!m_bOwnStrings)
            m_name=a->getNodeValue();
    }

    if (m_bOwnStrings) {
        XMLString::release(const_cast<XMLCh**>(&m_name));
        m_name=XMLString::replicate(name);
    }
}

void SAMLAttributeDesignator::setNamespace(const XMLCh* ns)
{
    if (!ns || !*ns)
        throw MalformedException("namespace cannot be null");
    
    if (m_root) {
        DOMAttr* a=static_cast<DOMElement*>(m_root)->getAttributeNodeNS(NULL,L(AttributeNamespace));
        a->setNodeValue(ns);
        if (!m_bOwnStrings)
            m_namespace=a->getNodeValue();
    }

    if (m_bOwnStrings) {
        XMLString::release(const_cast<XMLCh**>(&m_namespace));
        m_name=XMLString::replicate(ns);
    }
}

DOMNode* SAMLAttributeDesignator::toDOM(DOMDocument* doc, bool xmlns) const
{
    SAMLObject::toDOM(doc,xmlns);
    if (m_root)
    {
        if (xmlns)
            static_cast<DOMElement*>(m_root)->setAttributeNS(XML::XMLNS_NS,L(xmlns),XML::SAML_NS);
        return m_root;
    }
    if (!doc)
        doc=m_document;

    // Construct the Attribute.
    DOMElement* a=doc->createElementNS(XML::SAML_NS,L(AttributeDesignator));
    if (xmlns)
        a->setAttributeNS(XML::XMLNS_NS,L(xmlns),XML::SAML_NS);
    a->setAttributeNS(NULL,L(AttributeName),m_name);
    a->setAttributeNS(NULL,L(AttributeNamespace),m_namespace);

    return m_root=a;
}
