/*
 * The OpenSAML License, Version 1.
 * Copyright (c) 2002
 * University Corporation for Advanced Internet Development, Inc.
 * All rights reserved
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution, if any, must include
 * the following acknowledgment: "This product includes software developed by
 * the University Corporation for Advanced Internet Development
 * <http://www.ucaid.edu>Internet2 Project. Alternately, this acknowledegement
 * may appear in the software itself, if and wherever such third-party
 * acknowledgments normally appear.
 *
 * Neither the name of OpenSAML nor the names of its contributors, nor
 * Internet2, nor the University Corporation for Advanced Internet Development,
 * Inc., nor UCAID may be used to endorse or promote products derived from this
 * software without specific prior written permission. For written permission,
 * please contact opensaml@opensaml.org
 *
 * Products derived from this software may not be called OpenSAML, Internet2,
 * UCAID, or the University Corporation for Advanced Internet Development, nor
 * may OpenSAML appear in their name, without prior written permission of the
 * University Corporation for Advanced Internet Development.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND WITH ALL FAULTS. ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, AND NON-INFRINGEMENT ARE DISCLAIMED AND THE ENTIRE RISK
 * OF SATISFACTORY QUALITY, PERFORMANCE, ACCURACY, AND EFFORT IS WITH LICENSEE.
 * IN NO EVENT SHALL THE COPYRIGHT OWNER, CONTRIBUTORS OR THE UNIVERSITY
 * CORPORATION FOR ADVANCED INTERNET DEVELOPMENT, INC. BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/* SAMLSOAPBinding.cpp - basic SAML SOAP binding implementation

   Scott Cantor
   6/3/02

   $History:$
*/

#include "internal.h"
using namespace saml;

#include <ctime>
#include <sstream>
#include <list>
using namespace std;

#include <curl/curl.h>
#include <log4cpp/Category.hh>
#include <xercesc/framework/MemBufInputSource.hpp>
#include <xercesc/util/Base64.hpp>

// callback to buffer data from server
size_t curl_write_hook(void* ptr, size_t size, size_t nmemb, void* stream)
{
    // *stream is actually an ostream object
    ostream& buf=*(reinterpret_cast<ostream*>(stream));
    buf.write(reinterpret_cast<const char*>(ptr),size*nmemb);
    return size*nmemb;
}

// callback for curl debug data
int curl_debug_hook(CURL* handle, curl_infotype type, char* data, size_t len, void* ptr)
{
    // *log is actually a logging object
    Category& log=*(reinterpret_cast<Category*>(ptr));
    log.debug(string(data,(data[len-1]=='\n') ? len-1 : len));
    return 0;
}

// callback to invoke a caller-defined SSL callback, used because OpenSSL < 0.9.7 has no data ptr
CURLcode saml_ssl_ctx_callback(CURL* curl, void* ssl_ctx, void* userptr)
{
    SAMLConfig::SAMLBindingConfig* conf = reinterpret_cast<SAMLConfig::SAMLBindingConfig*>(userptr);
    if (conf->ssl_ctx_callback(ssl_ctx,conf->ssl_ctx_data))
        return CURLE_OK;
    return CURLE_SSL_CERTPROBLEM;
}

class CURLPool
{
public:
    CURLPool() : m_size(0), m_headers(NULL), m_log(&Category::getInstance(SAML_LOGCAT".SAMLSOAPBinding.CURLPool")) {
        m_headers=curl_slist_append(m_headers,"Content-Type: text/xml");
        m_headers=curl_slist_append(m_headers,"SOAPAction: http://www.oasis-open.org/committees/security");
        m_headers=curl_slist_append(m_headers,SAMLSOAPBinding::version.c_str());
    }
    ~CURLPool();
    
    CURL* get(const char* location);
    void put(const char* location, CURL* handle);
    typedef map<string,vector<CURL*> > poolmap_t;

private:    
    poolmap_t m_bindingMap;
    list< vector<CURL*>* > m_pools;
    long m_size;
    struct curl_slist* m_headers;
    Category* m_log;
};

CURLPool::~CURLPool()
{
    for (poolmap_t::iterator i=m_bindingMap.begin(); i!=m_bindingMap.end(); i++)
    {
        for (vector<CURL*>::iterator j=i->second.begin(); j!=i->second.end(); j++)
            curl_easy_cleanup(*j);
    }
    curl_slist_free_all(m_headers);
}

CURL* CURLPool::get(const char* location)
{
    saml::NDC("get");
    m_log->debug("getting connection handle to %s", location);
    SAMLConfig::getConfig().saml_lock();
    poolmap_t::iterator i=m_bindingMap.find(location);
    
    if (i!=m_bindingMap.end())
    {
        // Move this pool to the front of the list.
        m_pools.remove(&(i->second));
        m_pools.push_front(&(i->second));
        
        // If a free connection exists, return it.
        if (!(i->second.empty()))
        {
            CURL* handle=i->second.back();
            i->second.pop_back();
            m_size--;
            SAMLConfig::getConfig().saml_unlock();
            m_log->debug("returning existing connection handle from pool");
            return handle;
        }
    }
    
    SAMLConfig::getConfig().saml_unlock();
    m_log->debug("nothing free in pool, returning new connection handle");
    
    // Create a new connection and set non-varying options.
    CURL* handle=curl_easy_init();
    if (!handle)
        return NULL;
    curl_easy_setopt(handle,CURLOPT_NOPROGRESS,1);
    curl_easy_setopt(handle,CURLOPT_NOSIGNAL,1);
    curl_easy_setopt(handle,CURLOPT_FAILONERROR,1);
    curl_easy_setopt(handle,CURLOPT_SSLVERSION,3);
    curl_easy_setopt(handle,CURLOPT_SSL_VERIFYHOST,2);
    curl_easy_setopt(handle,CURLOPT_WRITEFUNCTION,&curl_write_hook);
    curl_easy_setopt(handle,CURLOPT_DEBUGFUNCTION,&curl_debug_hook);
    curl_easy_setopt(handle,CURLOPT_HTTPHEADER,m_headers);

    return handle;
}

void CURLPool::put(const char* location, CURL* handle)
{

    SAMLConfig::getConfig().saml_lock();

    poolmap_t::iterator i=m_bindingMap.find(location);
    if (i==m_bindingMap.end())
        m_pools.push_front(&(m_bindingMap.insert(poolmap_t::value_type(location,vector<CURL*>(1,handle))).first->second));
    else
        i->second.push_back(handle);
    
    CURL* killit=NULL;
    if (++m_size > SAMLConfig::getConfig().conn_pool_max)
    {
        // Kick a handle out from the back of the bus.
        while (true)
        {
            vector<CURL*>* corpse=m_pools.back();
            if (!corpse->empty())
            {
                killit=corpse->back();
                corpse->pop_back();
                m_size--;
                break;
            }
            
            // Move an empty pool up to the front so we don't keep hitting it.
            m_pools.pop_back();
            m_pools.push_front(corpse);
        }
    }
    SAMLConfig::getConfig().saml_unlock();
    if (killit) {
        curl_easy_cleanup(killit);
        saml::NDC("put");
        m_log->info("conn_pool_max limit reached, dropping an old connection");
    }
}

namespace {
   CURLPool* g_CURLPool=NULL;
}

void saml::soap_pool_init()
{
    g_CURLPool=new CURLPool();
}

void saml::soap_pool_term()
{
    delete g_CURLPool;
    g_CURLPool = NULL;
}

string SAMLSOAPBinding::version;

SAMLSOAPBinding::SAMLSOAPBinding() {}

SAMLSOAPBinding::~SAMLSOAPBinding() {}

#define CHECK_CURL_RESP(c) \
    if (c!=CURLE_OK) { \
        log.error("failed while contacting SAML responder: %s", \
            (curl_errorbuf[0] ? curl_errorbuf : "no further information available")); \
        throw BindingException(SAMLException::RESPONDER, \
            string("SAMLSOAPBinding::send() failed while contacting SAML responder: ") + \
                (curl_errorbuf[0] ? curl_errorbuf : "no further information available")); \
    }

SAMLResponse* SAMLSOAPBinding::send(const SAMLAuthorityBinding& bindingInfo, SAMLRequest& req, SAMLConfig::SAMLBindingConfig& conf)
{
    NDC ndc("send");
    Category& log=Category::getInstance(SAML_LOGCAT".SAMLSOAPBinding");
    Category& log_curl=Category::getInstance(SAML_LOGCAT".libcurl");

    // Turn the request into a DOM, and use its document for the SOAP nodes.
    DOMDocument* doc=req.toDOM()->getOwnerDocument();

    // Build a SOAP envelope and body.
    DOMElement* e=doc->createElementNS(XML::SOAP11ENV_NS,L(Envelope));
    e->setAttributeNS(XML::XMLNS_NS,L(xmlns),XML::SOAP11ENV_NS);
    DOMElement* body=doc->createElementNS(XML::SOAP11ENV_NS,L(Body));
    e->appendChild(body);

    // Attach SAML request.
    body->appendChild(req.toDOM());

    if (!doc->getDocumentElement())
        doc->appendChild(e);
    else
        doc->replaceChild(e, doc->getDocumentElement());

    // Serialize the DOM.
    stringstream inbuf;
    ostringstream outbuf;
    string outbufstring;
    inbuf << *(doc->getDocumentElement());

    auto_ptr_char location(bindingInfo.getLocation());
    CURL* handle=g_CURLPool->get(location.get());
    if (!handle)
        throw SAMLException(SAMLException::REQUESTER,"SAMLSOAPBinding::send() unable to obtain a curl handle");
    
    string inbufstring=inbuf.str();
    
    // Setup the per-call curl properties.
    curl_easy_setopt(handle,CURLOPT_URL,location.get());
    curl_easy_setopt(handle,CURLOPT_POSTFIELDS,inbufstring.c_str());
    curl_easy_setopt(handle,CURLOPT_POSTFIELDSIZE,inbufstring.length());
    curl_easy_setopt(handle,CURLOPT_FILE,&outbuf);
    curl_easy_setopt(handle,CURLOPT_DEBUGDATA,&log_curl);

    char curl_errorbuf[CURL_ERROR_SIZE];
    curl_errorbuf[0]=0;
    curl_easy_setopt(handle,CURLOPT_ERRORBUFFER,curl_errorbuf);
    if (log.isDebugEnabled())
        curl_easy_setopt(handle,CURLOPT_VERBOSE,1);

    if (conf.timeout>=0)
        curl_easy_setopt(handle,CURLOPT_TIMEOUT,conf.timeout);
    if (conf.conn_timeout>=0)
        curl_easy_setopt(handle,CURLOPT_CONNECTTIMEOUT,conf.conn_timeout);

    if (!conf.ssl_calist.empty())
        curl_easy_setopt(handle,CURLOPT_CAINFO,conf.ssl_calist.c_str());
    else
        curl_easy_setopt(handle,CURLOPT_CAINFO,NULL);
    
    curl_easy_setopt(handle,CURLOPT_SSL_VERIFYPEER,conf.ssl_calist.empty() ? 0 : 1);

    if (!conf.ssl_certfile.empty() && !conf.ssl_keyfile.empty())
    {
        curl_easy_setopt(handle,CURLOPT_SSLCERT,conf.ssl_certfile.c_str());
        if (!conf.ssl_certtype.empty())
            curl_easy_setopt(handle,CURLOPT_SSLCERTTYPE,conf.ssl_certtype.c_str());
        curl_easy_setopt(handle,CURLOPT_SSLKEY,conf.ssl_keyfile.c_str());
        curl_easy_setopt(handle,CURLOPT_SSLKEYPASSWD,conf.ssl_keypass.c_str());
        if (!conf.ssl_keytype.empty())
            curl_easy_setopt(handle,CURLOPT_SSLKEYTYPE,conf.ssl_keytype.c_str());
    }
    else
    {
        curl_easy_setopt(handle,CURLOPT_SSLCERT,NULL);
        curl_easy_setopt(handle,CURLOPT_SSLCERTTYPE,NULL);
        curl_easy_setopt(handle,CURLOPT_SSLKEY,NULL);
        curl_easy_setopt(handle,CURLOPT_SSLKEYPASSWD,NULL);
        curl_easy_setopt(handle,CURLOPT_SSLKEYTYPE,NULL);
    }

    if (conf.ssl_ctx_callback)
    {
        curl_easy_setopt(handle,CURLOPT_SSL_CTX_FUNCTION,saml_ssl_ctx_callback);
        curl_easy_setopt(handle,CURLOPT_SSL_CTX_DATA,&conf);
    }
    else
    {
        curl_easy_setopt(handle,CURLOPT_SSL_CTX_FUNCTION,NULL);
        curl_easy_setopt(handle,CURLOPT_SSL_CTX_DATA,NULL);
    }

    // Make the call.
    try
    {
        log.info("sending SOAP message to %s", location.get());
        CURLcode cc=curl_easy_perform(handle);
        CHECK_CURL_RESP(cc);
    
        outbufstring=outbuf.str();
    
        // Interrogate the response.
        long http_status;
        cc=curl_easy_getinfo(handle,CURLINFO_HTTP_CODE,&http_status);
        CHECK_CURL_RESP(cc);
        char* content_type;
        cc=curl_easy_getinfo(handle,CURLINFO_CONTENT_TYPE,&content_type);
        CHECK_CURL_RESP(cc);
        
        if (!content_type || !strstr(content_type,"text/xml"))
        {
            unsigned int len;
            XMLByte* b=Base64::encode(const_cast<XMLByte*>((XMLByte*)outbufstring.c_str()),outbufstring.length(),&len);
            ContentTypeException exc(SAMLException::RESPONDER,reinterpret_cast<char*>(b));
            delete[] b;
            throw exc;
        }

        curl_easy_setopt(handle,CURLOPT_ERRORBUFFER,NULL);
        g_CURLPool->put(location.get(),handle);
    }
    catch (...)
    {
        curl_easy_setopt(handle,CURLOPT_ERRORBUFFER,NULL);
        curl_easy_cleanup(handle);
        throw;
    }
    
    // In theory, the SOAP message should be in buf. Get a parser and build the DOM.
    XML::Parser p;
    static const XMLCh systemId[]={chLatin_A, chLatin_A, chLatin_R, chLatin_e, chLatin_s, chLatin_p, chLatin_o, chLatin_n, chLatin_s, chLatin_e, chNull};
    MemBufInputSource membufsrc(reinterpret_cast<const XMLByte*>(outbufstring.c_str()),outbufstring.length(),systemId,false);
    Wrapper4InputSource dsrc(&membufsrc,false);
    DOMDocument* rdoc=p.parse(dsrc);

    try
    {
        DOMElement* e=rdoc->getDocumentElement();

        // The root must be a SOAP 1.1 envelope.
        if (XMLString::compareString(XML::SOAP11ENV_NS,e->getNamespaceURI()) ||
            XMLString::compareString(L(Envelope),e->getLocalName()))
            throw MalformedException(SAMLException::RESPONDER,"SAMLSOAPBinding::send() detected an incompatible or missing SOAP envelope");

        /* Walk the children. If we encounter any headers with mustUnderstand, we have to bail.
           The thinking here is, we're not a "real" SOAP processor, but we have to emulate one that
           understands no headers. For now, assume we're the recipient.
         */
        DOMNode* n=e->getFirstChild();
        while (n && n->getNodeType()!=DOMNode::ELEMENT_NODE)
            n=n->getNextSibling();
        if (!XMLString::compareString(XML::SOAP11ENV_NS,n->getNamespaceURI()) &&
            !XMLString::compareString(L(Header),n->getLocalName()))
        {
            DOMNode* header=n->getFirstChild();
            while (header)
            {
                if (header->getNodeType()==DOMNode::ELEMENT_NODE &&
                    (static_cast<DOMElement*>(header))->getAttributeNS(XML::SOAP11ENV_NS,L(mustUnderstand)) &&
                    XMLString::parseInt((static_cast<DOMElement*>(header))->getAttributeNS(XML::SOAP11ENV_NS,L(mustUnderstand)))==1)
                    throw SOAPException(SAMLException::RESPONDER,"SAMLSOAPBinding::send() detected a mandatory SOAP header");
                header=header->getNextSibling();
            }    
            n=n->getNextSibling();   // advance to the Body element
            while (n && n->getNodeType()!=DOMNode::ELEMENT_NODE)
                n=n->getNextSibling();
        }

        if (n)
        {
            // Get the first (and only) child element of the Body.
            n=n->getFirstChild();
            while (n && n->getNodeType()!=DOMNode::ELEMENT_NODE)
                n=n->getNextSibling();
            if (n)
            {
                // Is it a fault?
                if (n->getNodeType()==DOMNode::ELEMENT_NODE &&
                    !XMLString::compareString(XML::SOAP11ENV_NS,n->getNamespaceURI()) &&
                    !XMLString::compareString(L(Fault),n->getLocalName()))
                {
                    // Find the faultstring element and use it in the message.
                    DOMNodeList* nlist=(static_cast<DOMElement*>(n))->getElementsByTagNameNS(NULL,L(faultstring));
                    if (nlist && nlist->getLength())
                    {
                        vector<saml::QName> codes;
                        auto_ptr_char msg(nlist->item(0)->getFirstChild()->getNodeValue());
                        nlist=(static_cast<DOMElement*>(n))->getElementsByTagNameNS(NULL,L(faultcode));
                        if (nlist && nlist->getLength())
                        {
                            auto_ptr<saml::QName> qcode(saml::QName::getQNameTextNode(static_cast<DOMText*>(nlist->item(0)->getFirstChild())));
                            codes.push_back(*qcode);
                        }
                        throw SOAPException(Iterator<saml::QName>(codes),msg.get());
                    }
                    else
                        throw SOAPException(SAMLException::RESPONDER,"SAMLSOAPBinding::send() detected a SOAP fault");
                }

                SAMLResponse* response=new SAMLResponse(static_cast<DOMElement*>(n));
                if (XMLString::compareString(response->getInResponseTo(),req.getId())) {
                    delete response;
                    throw BindingException(SAMLException::RESPONDER,"SAMLSOAPBinding::send() received a SAML response with the wrong InResponseTo value");
                }
                response->setDocument(rdoc);
                return response;
            }
        }
        throw SOAPException(SAMLException::RESPONDER,"SAMLSOAPBinding::send() unable to find a SAML response or fault in SOAP body");
    }
    catch(...)
    {
        rdoc->release();
        throw;
    }
}

SAMLRequest* SAMLSOAPBinding::receive(void* reqContext, const XMLCh** requester)
{
    throw SAMLException(SAMLException::RESPONDER,"SAMLSOAPBinding::receive() not yet implemented");
}

void SAMLSOAPBinding::respond(void* respContext, SAMLResponse* response, SAMLException* e)
{
    throw SAMLException(SAMLException::RESPONDER,"SAMLSOAPBinding::respond() not yet implemented");
}
